# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Misc utils"""

from collections.abc import Sequence
from typing import Type, get_type_hints

import numpy as np


def is_sequence(seq):
    """
    Returns True for lists, tuples, sets, arrays
    Return False for strings, dicts
    """
    if isinstance(seq, str):
        return False

    if isinstance(seq, (np.ndarray, set)):
        return True
    return isinstance(seq, Sequence)


def get_inherited_attrs(cls: Type, *private_attributes):
    """
    Get the attribute from the object and all its parents
    """

    # The extras are needed for annotated types like NDArray3
    retrieved_attributes = get_type_hints(cls, include_extras=True)
    retrieved_attributes = {attr: type for attr, type in retrieved_attributes.items() if not attr.startswith("_")}

    for private_attr in private_attributes:
        for parent in reversed(list(cls.__mro__)):
            attr_dict = retrieved_attributes.get(private_attr, {})
            attr_dict.update(getattr(parent, private_attr, {}))
            retrieved_attributes[private_attr] = attr_dict

    return retrieved_attributes


def array_equal_with_nan(array1: np.ndarray, array2: np.ndarray) -> bool:
    """Compare two structured arrays for equality, treating NaN values as equal.

    np.array_equal does not work with NaN values in structured arrays, so we need to compare column by column.
    related issue: https://github.com/numpy/numpy/issues/21539
    """
    if array1.dtype.names != array2.dtype.names:
        return False

    columns: Sequence[str] = array1.dtype.names
    for column in columns:
        column_dtype = array1.dtype[column]
        if np.issubdtype(column_dtype, np.str_):
            if not np.array_equal(array1[column], array2[column]):
                return False
            continue
        if not np.array_equal(array1[column], array2[column], equal_nan=True):
            return False
    return True



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from typing import Any

import dash_cytoscape as cyto
from dash import html

from power_grid_model_ds._core.visualizer.layout.colors import BACKGROUND_COLOR
from power_grid_model_ds._core.visualizer.layout.cytoscape_styling import DEFAULT_STYLESHEET

LAYOUT_OPTIONS = ["random", "circle", "concentric", "grid", "cose", "breadthfirst"]

_CYTO_INNER_STYLE = {"width": "100%", "height": "100%", "background-color": BACKGROUND_COLOR}
_CYTO_OUTER_STYLE = {"height": "80vh"}


def get_cytoscape_html(layout: str, elements: list[dict[str, Any]]) -> html.Div:
    """Get the Cytoscape HTML element"""
    return html.Div(
        cyto.Cytoscape(
            id="cytoscape-graph",
            layout={"name": layout},
            style=_CYTO_INNER_STYLE,
            elements=elements,
            stylesheet=DEFAULT_STYLESHEET,
        ),
        style=_CYTO_OUTER_STYLE,
    )



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Optional dependencies for the arrays module."""

try:
    # pylint: disable=unused-import
    import pandas
except ImportError:
    pandas = None



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Node data types"""

import numpy as np
from numpy.typing import NDArray

from power_grid_model_ds._core.model.constants import empty
from power_grid_model_ds._core.model.dtypes.id import Id
from power_grid_model_ds._core.model.enums.nodes import NodeType


class Node(Id):
    """Node data type"""

    u_rated: NDArray[np.float64]  # rated line-line voltage
    node_type: NDArray[np.int8]
    feeder_branch_id: NDArray[np.int32]  # branch id of the feeder
    feeder_node_id: NDArray[np.int32]  # node id of the first substation node

    _defaults = {
        "node_type": NodeType.UNSPECIFIED.value,
        "feeder_branch_id": empty,
        "feeder_node_id": empty,
    }



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Contains selectors for the Cytoscape stylesheet."""

from power_grid_model_ds._core.visualizer.layout.colors import CYTO_COLORS

NODE_SIZE = 100
BRANCH_WIDTH = 10

_BRANCH_STYLE = {
    "selector": "edge",
    "style": {
        "line-color": CYTO_COLORS["line"],
        "target-arrow-color": CYTO_COLORS["line"],
        "curve-style": "bezier",
        "target-arrow-shape": "triangle",
        "width": BRANCH_WIDTH,
    },
}
_NODE_STYLE = {
    "selector": "node",
    "style": {
        "label": "data(id)",
        "border-width": 5,
        "border-color": "black",
        "font-size": 25,
        "text-halign": "center",
        "text-valign": "center",
        "background-color": CYTO_COLORS["node"],
        "text-background-color": CYTO_COLORS["node"],
        "text-background-opacity": 1,
        "text-background-shape": "round-rectangle",
        "width": 75,
        "height": 75,
    },
}
_NODE_LARGE_ID_STYLE = {
    "selector": "node[id > 10000000]",
    "style": {"font-size": 15},
}
_SELECTED_NODE_STYLE = {
    "selector": "node:selected, node:active",
    "style": {"border-width": 5, "border-color": CYTO_COLORS["selected"]},
}

_SELECTED_BRANCH_STYLE = {
    "selector": "edge:selected, edge:active",
    "style": {"line-color": CYTO_COLORS["selected"], "target-arrow-color": CYTO_COLORS["selected"], "width": 10},
}


_SUBSTATION_NODE_STYLE = {
    "selector": "node[node_type = 1]",
    "style": {
        "label": "data(id)",
        "shape": "diamond",
        "background-color": CYTO_COLORS["substation_node"],
        "text-background-color": CYTO_COLORS["substation_node"],
        "width": NODE_SIZE * 1.2,
        "height": NODE_SIZE * 1.2,
        "color": "white",
    },
}
_TRANSFORMER_STYLE = {
    "selector": "edge[group = 'transformer']",
    "style": {"line-color": CYTO_COLORS["transformer"], "target-arrow-color": CYTO_COLORS["transformer"]},
}
_SELECTED_TRANSFORMER_STYLE = {
    "selector": "edge[group = 'transformer']:selected, edge[group = 'transformer']:active",
    "style": {
        "line-color": CYTO_COLORS["selected_transformer"],
        "target-arrow-color": CYTO_COLORS["selected_transformer"],
    },
}

_OPEN_BRANCH_STYLE = {
    "selector": "edge[from_status = 0], edge[to_status = 0]",
    "style": {
        "line-style": "dashed",
        "line-color": CYTO_COLORS["open_branch"],
        "target-arrow-color": CYTO_COLORS["open_branch"],
        "source-arrow-color": CYTO_COLORS["open_branch"],
    },
}
_OPEN_FROM_SIDE_BRANCH_STYLE = {
    "selector": "edge[from_status = 0]",
    "style": {
        "source-arrow-shape": "diamond",
        "source-arrow-fill": "hollow",
    },
}
_OPEN_TO_SIDE_BRANCH_STYLE = {
    "selector": "edge[to_status = 0]",
    "style": {
        "target-arrow-shape": "diamond",
        "target-arrow-fill": "hollow",
    },
}


DEFAULT_STYLESHEET = [
    _NODE_STYLE,
    _NODE_LARGE_ID_STYLE,
    _SUBSTATION_NODE_STYLE,
    _BRANCH_STYLE,
    _TRANSFORMER_STYLE,
    _SELECTED_NODE_STYLE,
    _SELECTED_BRANCH_STYLE,
    _SELECTED_TRANSFORMER_STYLE,
    _OPEN_BRANCH_STYLE,
    _OPEN_FROM_SIDE_BRANCH_STYLE,
    _OPEN_TO_SIDE_BRANCH_STYLE,
]



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""A set of helper functions that mimic numpy functions but are specifically designed for FancyArrays."""

from typing import TYPE_CHECKING, Union

import numpy as np

from power_grid_model_ds._core.utils.misc import array_equal_with_nan

if TYPE_CHECKING:
    from power_grid_model_ds._core.model.arrays.base.array import FancyArray


def concatenate(fancy_array: "FancyArray", *other_arrays: Union["FancyArray", np.ndarray]) -> "FancyArray":
    """Concatenate arrays."""
    np_arrays = [array if isinstance(array, np.ndarray) else array.data for array in other_arrays]
    try:
        concatenated = np.concatenate([fancy_array.data] + np_arrays)
    except TypeError as error:
        raise TypeError("Cannot append arrays: mismatching dtypes.") from error
    return fancy_array.__class__(data=concatenated)


def unique(array: "FancyArray", **kwargs):
    """Return the unique elements of the array."""
    for column in array.columns:
        if np.issubdtype(array.dtype[column], np.floating) and np.isnan(array[column]).any():
            raise NotImplementedError("Finding unique records in array with NaN values is not supported.")
            # see https://github.com/numpy/numpy/issues/23286
    unique_data = np.unique(array.data, **kwargs)
    if isinstance(unique_data, tuple):
        unique_data, *other = unique_data
        return array.__class__(data=unique_data), *other
    return array.__class__(data=unique_data)


def sort(array: "FancyArray", axis=-1, kind=None, order=None) -> "FancyArray":
    """Sort the array in-place and return sorted array."""
    array.data.sort(axis=axis, kind=kind, order=order)
    return array


def array_equal(array1: "FancyArray", array2: "FancyArray", equal_nan: bool = True) -> bool:
    """Return True if two arrays are equal."""
    if equal_nan:
        return array_equal_with_nan(array1.data, array2.data)
    return np.array_equal(array1.data, array2.data)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import dash_bootstrap_components as dbc
from dash import html

from power_grid_model_ds._core.visualizer.layout.colors import CYTO_COLORS

_MARGIN = "0 10px"
_FONT_SIZE = "2.5em"

NODE_ICON_STYLE = {"font-size": _FONT_SIZE, "margin": _MARGIN, "color": CYTO_COLORS["node"]}
_SUBSTATION_ICON_STYLE = {"font-size": _FONT_SIZE, "margin": _MARGIN, "color": CYTO_COLORS["substation_node"]}
_LINE_ICON_STYLE = {"font-size": _FONT_SIZE, "margin": _MARGIN, "color": CYTO_COLORS["line"]}
_LINK_ICON_STYLE = {"font-size": _FONT_SIZE, "margin": _MARGIN, "color": CYTO_COLORS["link"]}
_TRANSFORMER_ICON_STYLE = {"font-size": _FONT_SIZE, "margin": _MARGIN, "color": CYTO_COLORS["transformer"]}
_OPEN_BRANCH_ICON_STYLE = {"font-size": _FONT_SIZE, "margin": _MARGIN, "color": CYTO_COLORS["open_branch"]}
LEGENDA_HTML = html.Div(
    [
        html.I(className="fas fa-circle", id="node-icon", style=NODE_ICON_STYLE),
        dbc.Tooltip("Node", target="node-icon", placement="bottom"),
        html.I(className="fas fa-diamond", id="substation-icon", style=_SUBSTATION_ICON_STYLE),
        dbc.Tooltip("Substation", target="substation-icon", placement="bottom"),
        html.I(className="fas fa-arrow-right-long", id="line-icon", style=_LINE_ICON_STYLE),
        dbc.Tooltip("Line", target="line-icon", placement="bottom"),
        html.I(className="fas fa-arrow-right-long", id="transformer-icon", style=_TRANSFORMER_ICON_STYLE),
        dbc.Tooltip("Transformer", target="transformer-icon", placement="bottom"),
        html.I(className="fas fa-arrow-right-long", id="link-icon", style=_LINK_ICON_STYLE),
        dbc.Tooltip("Link", target="link-icon", placement="bottom"),
        html.I(className="fas fa-ellipsis", id="open-branch-icon", style=_OPEN_BRANCH_ICON_STYLE),
        dbc.Tooltip("Open Branch", target="open-branch-icon", placement="bottom"),
    ],
    style={
        "display": "flex",
        "align-items": "center",
        "margin": _MARGIN,
        "width": "100%",
        "text-shadow": "0 0 5px #000",
    },
)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import dash_bootstrap_components as dbc
from dash import Dash, dcc, html
from dash_bootstrap_components.icons import FONT_AWESOME

from power_grid_model_ds._core.model.grids.base import Grid
from power_grid_model_ds._core.visualizer.callbacks import (  # noqa: F401  # pylint: disable=unused-import
    element_scaling,
    element_selection,
    layout_dropdown,
    search_form,
)
from power_grid_model_ds._core.visualizer.layout.cytoscape_html import get_cytoscape_html
from power_grid_model_ds._core.visualizer.layout.header import HEADER_HTML
from power_grid_model_ds._core.visualizer.layout.selection_output import SELECTION_OUTPUT_HTML
from power_grid_model_ds._core.visualizer.parsers import parse_branches, parse_node_array
from power_grid_model_ds.arrays import NodeArray

GOOGLE_FONTS = "https://fonts.googleapis.com/css?family=Roboto:300,400,500,700&display=swap"
MDBOOTSTRAP = "https://cdnjs.cloudflare.com/ajax/libs/mdb-ui-kit/8.2.0/mdb.min.css"


def visualize(grid: Grid, debug: bool = False, port: int = 8050) -> None:
    """Visualize the Grid.

    grid: Grid
        The grid to visualize.

    layout: str
        The layout to use.

        If 'layout' is not provided (""):
            And grid.node contains "x" and "y" columns:
                The layout will be set to "preset" which uses the x and y coordinates to place the nodes.
            Otherwise:
                The layout will be set to "breadthfirst", which is a hierarchical breadth-first-search (BFS) layout.
        Other options:
            - "random": A layout that places the nodes randomly.
            - "circle": A layout that places the nodes in a circle.
            - "concentric": A layout that places the nodes in concentric circles.
            - "grid": A layout that places the nodes in a grid matrix.
            - "cose": A layout that uses the CompoundSpring Embedder algorithm (force-directed layout)
    """

    app = Dash(
        external_stylesheets=[dbc.themes.BOOTSTRAP, dbc.icons.BOOTSTRAP, MDBOOTSTRAP, FONT_AWESOME, GOOGLE_FONTS]
    )
    app.layout = get_app_layout(grid)
    app.run(debug=debug, port=port)


def _get_columns_store(grid: Grid) -> dcc.Store:
    return dcc.Store(
        id="columns-store",
        data={
            "node": grid.node.columns,
            "line": grid.line.columns,
            "link": grid.link.columns,
            "transformer": grid.transformer.columns,
            "three_winding_transformer": grid.three_winding_transformer.columns,
            "branch": grid.branches.columns,
        },
    )


def get_app_layout(grid: Grid) -> html.Div:
    """Get the app layout."""
    columns_store = _get_columns_store(grid)
    graph_layout = _get_graph_layout(grid.node)
    elements = parse_node_array(grid.node) + parse_branches(grid)
    cytoscape_html = get_cytoscape_html(graph_layout, elements)

    return html.Div(
        [
            columns_store,
            HEADER_HTML,
            html.Hr(style={"border-color": "white", "margin": "0"}),
            cytoscape_html,
            SELECTION_OUTPUT_HTML,
        ],
    )


def _get_graph_layout(nodes: NodeArray) -> str:
    """Determine the graph layout"""
    if "x" in nodes.columns and "y" in nodes.columns:
        return "preset"
    return "breadthfirst"



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from dash import dcc, html

from power_grid_model_ds._core.visualizer.layout.colors import CYTO_COLORS
from power_grid_model_ds._core.visualizer.layout.cytoscape_html import LAYOUT_OPTIONS

NODE_SCALE_HTML = [
    html.I(className="fas fa-circle", style={"color": CYTO_COLORS["node"], "margin-right": "10px"}),
    dcc.Input(
        id="node-scale-input",
        type="number",
        value=1,
        min=0.1,
        step=0.1,
        style={"width": "75px"},
    ),
    html.Span(style={"margin-right": "10px"}),
]

EDGE_SCALE_HTML = [
    html.I(className="fas fa-arrow-right-long", style={"color": CYTO_COLORS["line"], "margin-right": "10px"}),
    dcc.Input(
        id="edge-scale-input",
        type="number",
        value=1,
        min=0.1,
        step=0.1,
        style={"width": "75px"},
    ),
]

SCALE_INPUTS = [
    html.Div(
        NODE_SCALE_HTML + EDGE_SCALE_HTML,
        style={"margin": "0 20px 0 10px"},
    ),
]

LAYOUT_DROPDOWN_HTML = [
    html.Div(
        dcc.Dropdown(
            id="dropdown-update-layout",
            placeholder="Select layout",
            value="",
            clearable=False,
            options=[{"label": name.capitalize(), "value": name} for name in LAYOUT_OPTIONS],
            style={"width": "200px"},
        ),
        style={"margin": "0 20px 0 10px"},
    )
]



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from power_grid_model_ds._core.load_flow import PowerGridModelInterface
from power_grid_model_ds._core.model.graphs.container import GraphContainer
from power_grid_model_ds._core.model.grids.base import Grid

__all__ = ["Grid", "GraphContainer", "PowerGridModelInterface"]



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

YELLOW = "#facc37"
CYTO_COLORS = {
    "line": YELLOW,
    "link": "green",
    "transformer": "#4290f5",
    "node": YELLOW,
    "selected": "#e28743",
    "selected_transformer": "#0349a3",
    "substation_node": "purple",
    "open_branch": "#c9c9c9",
    "highlighted": "#a10000",
}
BACKGROUND_COLOR = "#555555"



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import dash_bootstrap_components as dbc
from dash import html

SPAN_TEXT_STYLE = {"color": "white", "margin-right": "8px", "font-weight": "bold", "text-shadow": "0 0 5px #000"}
_INPUT_STYLE = {"width": "150px", "display": "inline-block"}
# Create your form components
GROUP_INPUT = dbc.Select(
    id="search-form-group-input",
    options=[
        {"label": "node", "value": "node"},
        {"label": "line", "value": "line"},
        {"label": "link", "value": "link"},
        {"label": "transformer", "value": "transformer"},
        {"label": "branch", "value": "branch"},
    ],
    value="node",  # Default value
    style=_INPUT_STYLE,
)

COLUMN_INPUT = dbc.Select(
    id="search-form-column-input",
    options=[{"label": "id", "value": "id"}],
    value="id",  # Default value
    style=_INPUT_STYLE,
)

VALUE_INPUT = dbc.Input(id="search-form-value-input", placeholder="Enter value", type="text", style=_INPUT_STYLE)

OPERATOR_INPUT = dbc.Select(
    id="search-form-operator-input",
    options=[
        {"label": "=", "value": "="},
        {"label": "<", "value": "<"},
        {"label": ">", "value": ">"},
        {"label": "!=", "value": "!="},
    ],
    value="=",  # Default value
    style={"width": "60px", "display": "inline-block", "margin": "0 8px"},
)


# Arrange as a sentence
SEARCH_FORM_HTML = html.Div(
    [
        html.Span("Search ", style=SPAN_TEXT_STYLE),
        GROUP_INPUT,
        html.Span(" with ", className="mx-2", style=SPAN_TEXT_STYLE),
        COLUMN_INPUT,
        OPERATOR_INPUT,
        VALUE_INPUT,
    ],
    style={
        "display": "flex",
        "align-items": "center",
        "justify-content": "center",  # Centers items horizontally
        "padding": "10px",
        "margin": "0 auto",  # Centers the container itself
        "width": "100%",  # Ensures the container takes full width
    },
)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0
from power_grid_model_ds._core.model.arrays.pgm_arrays import Branch3Array, BranchArray


def _get_branch3_branches(branch3: Branch3Array) -> BranchArray:
    node_1 = branch3.node_1.item()
    node_2 = branch3.node_2.item()
    node_3 = branch3.node_3.item()

    status_1 = branch3.status_1.item()
    status_2 = branch3.status_2.item()
    status_3 = branch3.status_3.item()

    branches = BranchArray.zeros(3)
    branches.from_node = [node_1, node_1, node_2]
    branches.to_node = [node_2, node_3, node_3]
    branches.from_status = [status_1, status_1, status_2]
    branches.to_status = [status_2, status_3, status_3]

    return branches



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import dash_bootstrap_components as dbc

from power_grid_model_ds._core.visualizer.layout.colors import BACKGROUND_COLOR
from power_grid_model_ds._core.visualizer.layout.cytoscape_config import LAYOUT_DROPDOWN_HTML, SCALE_INPUTS
from power_grid_model_ds._core.visualizer.layout.legenda import LEGENDA_HTML
from power_grid_model_ds._core.visualizer.layout.search_form import SEARCH_FORM_HTML

_SEARCH_FORM_CARD_STYLE = {
    "background-color": "#555555",
    "color": "white",
    "border-left": "1px solid white",
    "border-right": "1px solid white",
    "border-radius": 0,
}


HEADER_HTML = dbc.Row(
    [
        dbc.Col(LEGENDA_HTML, className="d-flex align-items-center"),
        dbc.Col(
            dbc.Card(SEARCH_FORM_HTML, style=_SEARCH_FORM_CARD_STYLE),
            className="d-flex justify-content-center align-items-center",
        ),
        dbc.Col(SCALE_INPUTS + LAYOUT_DROPDOWN_HTML, className="d-flex justify-content-end align-items-center"),
    ],
    style={"background-color": BACKGROUND_COLOR},
)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Model constants"""

import numpy as np


def empty(dtype: type):
    """Returns the empty value for the given dtype."""
    if hasattr(dtype, "subdtype") and hasattr(dtype, "base"):  # special case for custom 'NDArray3' dtype
        dtype = dtype.base

    if np.issubdtype(dtype, np.str_):
        return ""
    if np.issubdtype(dtype, np.dtype("bool")):
        return False
    if np.issubdtype(dtype, np.floating):
        return np.nan
    try:
        return np.iinfo(dtype).min
    except ValueError as error:
        raise ValueError("Unsupported dtype") from error


EMPTY_ID = empty(np.int32)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from dash import Input, Output, callback


@callback(Output("cytoscape-graph", "layout"), Input("dropdown-update-layout", "value"), prevent_initial_call=True)
def update_layout(layout):
    """Callback to update the layout of the graph."""
    return {"name": layout, "animate": True}



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from copy import deepcopy
from typing import Any

from dash import Input, Output, callback

from power_grid_model_ds._core.visualizer.layout.cytoscape_styling import BRANCH_WIDTH, DEFAULT_STYLESHEET, NODE_SIZE


@callback(
    Output("cytoscape-graph", "stylesheet", allow_duplicate=True),
    Input("node-scale-input", "value"),
    Input("edge-scale-input", "value"),
    prevent_initial_call=True,
)
def scale_elements(node_scale: float, edge_scale: float) -> list[dict[str, Any]]:
    """Callback to scale the elements of the graph."""
    new_stylesheet = deepcopy(DEFAULT_STYLESHEET)
    edge_style = {
        "selector": "edge",
        "style": {
            "width": BRANCH_WIDTH * edge_scale,
        },
    }
    new_stylesheet.append(edge_style)
    node_style = {
        "selector": "node",
        "style": {
            "height": NODE_SIZE * node_scale,
            "width": NODE_SIZE * node_scale,
        },
    }
    new_stylesheet.append(node_style)
    return new_stylesheet



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from typing import Any, Iterable, Literal

import numpy as np
from numpy.typing import NDArray

from power_grid_model_ds._core.model.arrays.base.errors import MultipleRecordsReturned, RecordDoesNotExist
from power_grid_model_ds._core.utils.misc import is_sequence


def get_filter_mask(
    *args: int | Iterable[int] | np.ndarray,
    array: np.ndarray,
    mode_: Literal["AND", "OR"],
    **kwargs: Any | list[Any] | np.ndarray,
) -> np.ndarray:
    """Returns a mask that matches the input parameters."""
    parsed_kwargs = _parse(args, kwargs)

    if invalid_kwargs := set(parsed_kwargs.keys()) - set(array.dtype.names or ()):
        raise ValueError(f"Invalid kwargs: {invalid_kwargs}")

    filter_mask = _initialize_filter_mask(mode_, array.size)
    for field, values in parsed_kwargs.items():
        field_mask = _build_filter_mask_for_field(array, field, values)
        if mode_ == "AND":
            filter_mask &= field_mask
        elif mode_ == "OR":
            filter_mask |= field_mask
        else:
            raise ValueError(f"Invalid mode: {mode_}, must be 'AND' or 'OR'")
    return filter_mask


def apply_filter(
    *args: int | Iterable[int] | np.ndarray,
    array: np.ndarray,
    mode_: Literal["AND", "OR"],
    **kwargs: Any | list[Any] | np.ndarray,
) -> np.ndarray:
    """Return an array with the records that match the input parameters.
    Note: output could be an empty array."""
    filter_mask = get_filter_mask(*args, array=array, mode_=mode_, **kwargs)
    return array[filter_mask]


def apply_exclude(
    *args: int | Iterable[int] | np.ndarray,
    array: np.ndarray,
    mode_: Literal["AND", "OR"],
    **kwargs: Any | list[Any] | np.ndarray,
) -> np.ndarray:
    """Return an array without records that match the input parameters.
    Note: output could be an empty array."""
    filter_mask = get_filter_mask(*args, array=array, mode_=mode_, **kwargs)
    return array[~filter_mask]


def apply_get(
    *args: int | Iterable[int] | np.ndarray,
    array: np.ndarray,
    mode_: Literal["AND", "OR"],
    **kwargs: Any | list[Any] | np.ndarray,
) -> np.ndarray:
    """Returns a record that matches the input parameters.
    If no or multiple records match the input parameters, an error is raised.
    """
    filtered_array = apply_filter(*args, array=array, mode_=mode_, **kwargs)
    if filtered_array.size == 1:
        return filtered_array

    args_str = f"\n\twith args: {args[0]}" if args else ""
    kwargs_str = f"\n\twith kwargs: {kwargs}" if kwargs else ""
    if filtered_array.size == 0:
        raise RecordDoesNotExist(f"No record found! {args_str}{kwargs_str}")
    raise MultipleRecordsReturned(f"Found more than one record! {args_str}{kwargs_str}")


def _build_filter_mask_for_field(array: np.ndarray, field: str, values) -> np.ndarray:
    if not is_sequence(values):
        # Note: is_sequence() does not consider a string as a sequence.
        values = [values]

    if not len(values):  # pylint: disable=use-implicit-booleaness-not-len
        return np.full(array.size, False)
    if isinstance(values, set):
        values = list(values)
    if len(values) == 1:  # speed-up for single value
        return array[field] == values[0]
    return np.isin(array[field], values)


def _parse(args: tuple[int | Iterable[int] | NDArray, ...] | NDArray[np.int64], kwargs):
    if not args and not kwargs:
        raise TypeError("No input provided.")
    if len(args) > 1:
        raise ValueError("Cannot parse more than 1 positional argument.")
    if len(args) == 1 and "id" in kwargs:
        raise ValueError("Cannot parse both positional argument and keyword argument 'id'.")
    if len(args) == 1 and isinstance(args[0], int):
        kwargs.update({"id": args})
    elif len(args) == 1:
        kwargs.update({"id": args[0]})
    return kwargs


def _initialize_filter_mask(mode_: Literal["AND", "OR"], size: int) -> np.ndarray:
    if mode_ == "AND":
        return np.full(size, True)
    if mode_ == "OR":
        return np.full(size, False)
    raise ValueError(f"Invalid mode: {mode_}, must be 'AND' or 'OR'")



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Base grid classes"""

import dataclasses
import itertools
import logging
from copy import copy
from dataclasses import dataclass
from pathlib import Path
from typing import Type, TypeVar

import numpy as np
import numpy.typing as npt

from power_grid_model_ds._core import fancypy as fp
from power_grid_model_ds._core.model.arrays import (
    AsymVoltageSensorArray,
    Branch3Array,
    BranchArray,
    LineArray,
    LinkArray,
    NodeArray,
    SourceArray,
    SymGenArray,
    SymLoadArray,
    SymPowerSensorArray,
    SymVoltageSensorArray,
    ThreeWindingTransformerArray,
    TransformerArray,
    TransformerTapRegulatorArray,
)
from power_grid_model_ds._core.model.arrays.base.array import FancyArray
from power_grid_model_ds._core.model.arrays.base.errors import RecordDoesNotExist
from power_grid_model_ds._core.model.containers.base import FancyArrayContainer
from power_grid_model_ds._core.model.enums.nodes import NodeType
from power_grid_model_ds._core.model.graphs.container import GraphContainer
from power_grid_model_ds._core.model.graphs.models import RustworkxGraphModel
from power_grid_model_ds._core.model.graphs.models.base import BaseGraphModel
from power_grid_model_ds._core.model.grids._text_sources import TextSource
from power_grid_model_ds._core.model.grids.helpers import set_feeder_ids, set_is_feeder
from power_grid_model_ds._core.utils.pickle import get_pickle_path, load_from_pickle, save_to_pickle
from power_grid_model_ds._core.utils.zip import file2gzip

Self = TypeVar("Self", bound="Grid")

# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-public-methods


@dataclass
class Grid(FancyArrayContainer):
    """Grid object containing the entire network and interface to interact with it.

    Examples:

        >>> from power_grid_model_ds import Grid
        >>> grid = Grid.empty()
        >>> grid
    """

    graphs: GraphContainer
    """The graph representations of the grid."""

    # nodes
    node: NodeArray

    # branches
    transformer: TransformerArray
    three_winding_transformer: ThreeWindingTransformerArray
    line: LineArray
    link: LinkArray

    source: SourceArray
    sym_load: SymLoadArray
    sym_gen: SymGenArray

    # regulators
    transformer_tap_regulator: TransformerTapRegulatorArray

    # sensors
    sym_power_sensor: SymPowerSensorArray
    sym_voltage_sensor: SymVoltageSensorArray
    asym_voltage_sensor: AsymVoltageSensorArray

    def __str__(self) -> str:
        """String representation of the grid.

        Compatible with https://csacademy.com/app/graph_editor/
        """
        grid_str = ""

        for transformer3 in self.three_winding_transformer:
            nodes = [transformer3.node_1.item(), transformer3.node_2.item(), transformer3.node_3.item()]
            for combo in itertools.combinations(nodes, 2):
                grid_str += f"S{combo[0]} S{combo[1]} {transformer3.id.item()},3-transformer\n"

        for branch in self.branches:
            from_node = self.node.get(id=branch.from_node).record
            to_node = self.node.get(id=branch.to_node).record

            from_node_str = f"S{from_node.id}" if from_node.node_type == NodeType.SUBSTATION_NODE else str(from_node.id)
            to_node_str = f"S{to_node.id}" if to_node.node_type == NodeType.SUBSTATION_NODE else str(to_node.id)

            suffix_str = str(branch.id.item())
            if branch.from_status.item() == 0 or branch.to_status.item() == 0:
                suffix_str = f"{suffix_str},open"

            if branch.id in self.transformer.id:
                suffix_str = f"{suffix_str},transformer"
            elif branch.id in self.link.id:
                suffix_str = f"{suffix_str},link"
            elif branch.id in self.line.id:
                pass  # no suffix needed
            else:
                raise ValueError(f"Branch {branch.id} is not a transformer, link or line")

            grid_str += f"{from_node_str} {to_node_str} {suffix_str}\n"
        return grid_str

    @property
    def branches(self) -> BranchArray:
        """Converts all branch arrays into a single BranchArray."""
        branch_dtype = BranchArray.get_dtype()
        branches = BranchArray()
        for array in self.branch_arrays:
            new_branch = BranchArray(data=array.data[list(branch_dtype.names)])
            branches = fp.concatenate(branches, new_branch)
        return branches

    @property
    def branch_arrays(self) -> list[BranchArray]:
        """Returns all branch arrays"""
        branch_arrays = []
        for field in dataclasses.fields(self):
            array = getattr(self, field.name)
            if isinstance(array, BranchArray):
                branch_arrays.append(array)
        return branch_arrays

    def get_typed_branches(self, branch_ids: list[int] | npt.NDArray[np.int32]) -> BranchArray:
        """Find a matching LineArray, LinkArray or TransformerArray for the given branch_ids

        Raises:
            ValueError:
                - If no branch_ids are provided.
                - If not all branch_ids are of the same type.
        """
        if not np.any(branch_ids):
            raise ValueError("No branch_ids provided.")
        for branch_array in self.branch_arrays:
            array = branch_array.filter(branch_ids)
            if 0 < array.size != len(branch_ids):
                raise ValueError("Branches are not of the same type.")
            if array.size:
                return array
        raise RecordDoesNotExist(f"Branches {branch_ids} not found in grid.")

    def reverse_branches(self, branches: BranchArray):
        """Reverse the direction of the branches."""
        if not branches.size:
            return
        if not isinstance(branches, (LineArray, LinkArray, TransformerArray)):
            try:
                branches = self.get_typed_branches(branches.id)
            except ValueError:
                # If the branches are not of the same type, reverse them per type (though this is slower)
                for array in self.branch_arrays:
                    self.reverse_branches(array.filter(branches.id))
                return

        from_nodes = branches.from_node
        to_nodes = branches.to_node

        array_field = self.find_array_field(branches.__class__)
        array = getattr(self, array_field.name)
        array.update_by_id(branches.id, from_node=to_nodes, to_node=from_nodes)

    @classmethod
    def empty(cls: Type[Self], graph_model: type[BaseGraphModel] = RustworkxGraphModel) -> Self:
        """Create an empty grid

        Args:
            graph_model (type[BaseGraphModel], optional): The graph model to use. Defaults to RustworkxGraphModel.

        Returns:
            Grid: An empty grid
        """
        empty_fields = cls._get_empty_fields()
        empty_fields["graphs"] = GraphContainer.empty(graph_model=graph_model)
        return cls(**empty_fields)

    def append(self, array: FancyArray, check_max_id: bool = True):
        """Append an array to the grid. Both 'grid arrays' and 'grid.graphs' will be updated.

        Args:
            array (FancyArray): The array to append.
            check_max_id (bool, optional): Whether to check if the array id is the maximum id. Defaults to True.
        """
        self._append(array, check_max_id=check_max_id)  # noqa

        # pylint: disable=protected-access
        self.graphs._append(array)

    def add_branch(self, branch: BranchArray) -> None:
        """Add a branch to the grid

        Args:
            branch (BranchArray): The branch to add
        """
        self._append(array=branch)
        self.graphs.add_branch_array(branch_array=branch)

        logging.debug(f"added branch {branch.id} from {branch.from_node} to {branch.to_node}")

    def delete_branch(self, branch: BranchArray) -> None:
        """Remove a branch from the grid

        Args:
            branch (BranchArray): The branch to remove
        """
        _add_branch_array(branch=branch, grid=self)
        self.graphs.delete_branch(branch=branch)
        logging.debug(
            f"""deleted branch {branch.id.item()} from {branch.from_node.item()} to {branch.to_node.item()}"""
        )

    def delete_branch3(self, branch: Branch3Array) -> None:
        """Remove a branch3 from the grid

        Args:
            branch (Branch3Array): The branch3 to remove
        """
        _add_branch_array(branch=branch, grid=self)
        self.graphs.delete_branch3(branch=branch)

    def add_node(self, node: NodeArray) -> None:
        """Add a new node to the grid

        Args:
            node (NodeArray): The node to add
        """
        self._append(array=node)
        self.graphs.add_node_array(node_array=node)
        logging.debug(f"added rail {node.id}")

    def delete_node(self, node: NodeArray) -> None:
        """Remove a node from the grid

        Args:
            node (NodeArray): The node to remove
        """
        self.node = self.node.exclude(id=node.id)
        self.sym_load = self.sym_load.exclude(node=node.id)
        self.source = self.source.exclude(node=node.id)

        for branch_array in self.branch_arrays:
            matching_branches = branch_array.filter(from_node=node.id, to_node=node.id, mode_="OR")
            for branch in matching_branches:
                self.delete_branch(branch)

        self.graphs.delete_node(node=node)
        logging.debug(f"deleted rail {node.id}")

    def make_active(self, branch: BranchArray) -> None:
        """Make a branch active

        Args:
            branch (BranchArray): The branch to make active
        """
        array_field = self.find_array_field(branch.__class__)
        array_attr = getattr(self, array_field.name)
        branch_mask = array_attr.id == branch.id
        array_attr.from_status[branch_mask] = 1
        array_attr.to_status[branch_mask] = 1
        setattr(self, array_field.name, array_attr)

        self.graphs.make_active(branch=branch)
        logging.debug(f"activated branch {branch.id}")

    def make_inactive(self, branch: BranchArray, at_to_side: bool = True) -> None:
        """Make a branch inactive. This is done by setting from or to status to 0.

        Args:
            branch (BranchArray): The branch to make inactive
            at_to_side (bool, optional): Whether to deactivate the to_status instead of the from_status.
            Defaults to True.
        """
        array_field = self.find_array_field(branch.__class__)
        array_attr = getattr(self, array_field.name)
        branch_mask = array_attr.id == branch.id
        status_side = "to_status" if at_to_side else "from_status"
        array_attr[status_side][branch_mask] = 0
        setattr(self, array_field.name, array_attr)

        self.graphs.make_inactive(branch=branch)
        logging.debug(f"deactivated branch {branch.id}")

    def get_branches_in_path(self, nodes_in_path: list[int]) -> BranchArray:
        """Returns all branches within a path of nodes

        Args:
            nodes_in_path (list[int]): The nodes in the path

        Returns:
            BranchArray: The branches in the path
        """
        return self.branches.filter(from_node=nodes_in_path, to_node=nodes_in_path, from_status=1, to_status=1)

    def get_nearest_substation_node(self, node_id: int):
        """Find the nearest substation node.

        Args:
            node_id(int): The id of the node to find the nearest substation node for.

        Returns:
            NodeArray: The nearest substation node.

        Raises:
            RecordDoesNotExist: If no substation node is connected to the input node.
        """
        connected_nodes = self.graphs.active_graph.get_connected(node_id=node_id, inclusive=True)
        substation_nodes = self.node.filter(node_type=NodeType.SUBSTATION_NODE.value)

        for node in connected_nodes:
            if node in substation_nodes.id:
                return substation_nodes.get(node)
        raise RecordDoesNotExist(f"No {NodeType.SUBSTATION_NODE.name} connected to node {node_id}")

    def get_downstream_nodes(self, node_id: int, inclusive: bool = False):
        """Get the downstream nodes from a node.
        Assuming each node has a single feeding substation and the grid is radial

        Example:
            given this graph: [1] - [2] - [3] - [4], with 1 being a substation node

            >>> graph.get_downstream_nodes(2) == [3, 4]
            >>> graph.get_downstream_nodes(3) == [4]
            >>> graph.get_downstream_nodes(3, inclusive=True) == [3, 4]

        Args:
            node_id(int): The id of the node to get the downstream nodes from.
            inclusive(bool): Whether to include the input node in the result.

        Raises:
            NotImplementedError: If the input node is a substation node.

        Returns:
            list[int]: The downstream nodes.
        """
        substation_nodes = self.node.filter(node_type=NodeType.SUBSTATION_NODE.value)

        if node_id in substation_nodes.id:
            raise NotImplementedError("get_downstream_nodes is not implemented for substation nodes!")

        return self.graphs.active_graph.get_downstream_nodes(
            node_id=node_id, start_node_ids=list(substation_nodes.id), inclusive=inclusive
        )

    def cache(self, cache_dir: Path, cache_name: str, compress: bool = True):
        """Cache Grid to a folder

        Args:
            cache_dir (Path): The directory to save the cache to.
            cache_name (str): The name of the cache.
            compress (bool, optional): Whether to compress the cache. Defaults to True.
        """
        tmp_graphs = copy(self.graphs)
        self.graphs = None  # noqa
        cache_dir.mkdir(parents=True, exist_ok=True)

        pickle_path = cache_dir / f"{cache_name}.pickle"
        save_to_pickle(path=pickle_path, python_object=self)

        if compress:
            gzip_path = file2gzip(pickle_path)
            pickle_path.unlink()
            return gzip_path

        self.graphs = tmp_graphs
        return pickle_path

    @classmethod
    # pylint: disable=arguments-differ
    def from_cache(cls: Type[Self], cache_path: Path, load_graphs: bool = True) -> Self:
        """Read from cache and build .graphs from arrays

        Args:
            cache_path (Path): The path to the cache
            load_graphs (bool, optional): Whether to load the graphs. Defaults to True.

        Returns:
            Self: The grid loaded from cache
        """
        pickle_path = get_pickle_path(cache_path)

        grid = cls._from_pickle(pickle_path=pickle_path)
        if load_graphs:
            grid.graphs = GraphContainer.from_arrays(grid)
        return grid

    @classmethod
    def _from_pickle(cls, pickle_path: Path):
        grid = load_from_pickle(path=pickle_path)
        if not isinstance(grid, Grid):
            raise TypeError(f"{pickle_path.name} is not a valid {cls.__name__} cache.")
        return grid

    @classmethod
    def from_txt(cls, *args: str):
        """Build a grid from a list of strings

        See the documentation for the expected format of the txt_lines

        Args:
            *args (str): The lines of the grid

        Examples:
            >>> Grid.from_txt("1 2", "2 3", "3 4 transformer", "4 5", "S1 6")
            alternative: Grid.from_txt("1 2\n2 3\n3 4 transformer\n4 5\nS1 6")
        """
        return TextSource(grid_class=cls).load_from_txt(*args)

    @classmethod
    # pylint: disable=arguments-differ
    def from_txt_file(cls, txt_file_path: Path):
        """Load grid from txt file

        Args:
            txt_file_path (Path): The path to the txt file
        """
        with open(txt_file_path, "r", encoding="utf-8") as f:
            txt_lines = f.readlines()
        return TextSource(grid_class=cls).load_from_txt(*txt_lines)

    def set_feeder_ids(self):
        """Sets feeder and substation id properties in the grids arrays"""
        set_is_feeder(grid=self)
        set_feeder_ids(grid=self)

    @classmethod
    def from_extended(cls, extended: "Grid") -> "Grid":
        """Create a grid from an extended Grid object."""
        new_grid = cls.empty()

        # Add nodes first, so that branches can reference them
        new_grid.append(new_grid.node.__class__.from_extended(extended.node))

        for field in dataclasses.fields(cls):
            if field.name == "node":
                continue  # already added
            if issubclass(field.type, FancyArray):
                extended_array = getattr(extended, field.name)
                new_array = field.type.from_extended(extended_array)
                new_grid.append(new_array, check_max_id=False)

        return new_grid


def _add_branch_array(branch: BranchArray | Branch3Array, grid: Grid):
    """Add a branch array to the grid"""
    array_field = grid.find_array_field(branch.__class__)
    array_attr = getattr(grid, array_field.name)
    setattr(grid, array_field.name, array_attr.exclude(id=branch.id))

    grid.transformer_tap_regulator = grid.transformer_tap_regulator.exclude(regulated_object=branch.id)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from typing import Any, Literal

from power_grid_model_ds._core.model.arrays.base.array import FancyArray
from power_grid_model_ds._core.model.grids.base import Grid
from power_grid_model_ds._core.model.utils import _get_branch3_branches
from power_grid_model_ds.arrays import Branch3Array, BranchArray, NodeArray


def parse_node_array(nodes: NodeArray) -> list[dict[str, Any]]:
    """Parse the nodes."""
    parsed_nodes = []

    with_coords = "x" in nodes.columns and "y" in nodes.columns

    columns = nodes.columns
    for node in nodes:
        cyto_elements = {"data": _array_to_dict(node, columns)}
        cyto_elements["data"]["id"] = str(node.id.item())
        cyto_elements["data"]["group"] = "node"
        if with_coords:
            cyto_elements["position"] = {"x": node.x.item(), "y": -node.y.item()}  # invert y-axis for visualization
        parsed_nodes.append(cyto_elements)
    return parsed_nodes


def parse_branches(grid: Grid) -> list[dict[str, Any]]:
    """Parse the branches."""
    parsed_branches = []
    parsed_branches.extend(parse_branch_array(grid.line, "line"))
    parsed_branches.extend(parse_branch_array(grid.link, "link"))
    parsed_branches.extend(parse_branch_array(grid.transformer, "transformer"))
    parsed_branches.extend(parse_branch3_array(grid.three_winding_transformer, "transformer"))
    return parsed_branches


def parse_branch3_array(branches: Branch3Array, group: Literal["transformer"]) -> list[dict[str, Any]]:
    """Parse the three-winding transformer array."""
    parsed_branches = []
    columns = branches.columns
    for branch in branches:
        for branch_ in _get_branch3_branches(branch):
            cyto_elements = {"data": _array_to_dict(branch_, columns)}
            cyto_elements["data"].update(
                {
                    # IDs need to be unique, so we combine the branch ID with the from and to nodes
                    "id": str(branch.id.item()) + f"_{branch_.from_node.item()}_{branch_.to_node.item()}",
                    "source": str(branch_.from_node.item()),
                    "target": str(branch_.to_node.item()),
                    "group": group,
                }
            )
            parsed_branches.append(cyto_elements)
    return parsed_branches


def parse_branch_array(branches: BranchArray, group: Literal["line", "link", "transformer"]) -> list[dict[str, Any]]:
    """Parse the branch array."""
    parsed_branches = []
    columns = branches.columns
    for branch in branches:
        cyto_elements = {"data": _array_to_dict(branch, columns)}
        cyto_elements["data"].update(
            {
                "id": str(branch.id.item()),
                "source": str(branch.from_node.item()),
                "target": str(branch.to_node.item()),
                "group": group,
            }
        )
        parsed_branches.append(cyto_elements)
    return parsed_branches


def _array_to_dict(array_record: FancyArray, columns: list[str]) -> dict[str, Any]:
    """Stringify the record (required by Dash)."""
    return dict(zip(columns, array_record.tolist().pop()))



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Contains the build_array function."""

import logging
from collections.abc import Sized
from typing import Any, Iterable

import numpy as np

from power_grid_model_ds._core.model.constants import empty
from power_grid_model_ds._core.utils.misc import is_sequence


def build_array(*args: tuple[Any], dtype: np.dtype, defaults: dict[str, np.generic], **kwargs) -> np.ndarray:
    """Constructs the array from the given args/kwargs."""
    parsed_input, size = _parse_input(*args, dtype=dtype, **kwargs)

    array: np.ndarray = np.zeros(size, dtype=dtype)
    _fill_defaults(array, defaults)

    if not size:
        return array

    if isinstance(parsed_input, np.ndarray) and parsed_input.dtype.names:
        _check_missing_columns(array.dtype.names or (), defaults, set(parsed_input.dtype.names))
        return _parse_structured_array(parsed_input, array)
    if isinstance(parsed_input, np.ndarray):
        # Note: defaults are not supported when working with unstructured arrays
        return _parse_array(parsed_input, array.dtype)

    _check_missing_columns(array.dtype.names or (), defaults, set(parsed_input.keys()))
    _fill_with_kwargs(array, parsed_input)
    return array


def _parse_input(*args: Any, dtype: np.dtype, **kwargs):
    """Combines the args and kwargs to a dict."""
    columns: list[str] = list(dtype.names) if dtype.names else []
    if args and kwargs:
        raise TypeError("Cannot construct from both args and kwargs")

    if args and isinstance(args[0], np.ndarray):
        return args[0], len(args[0])
    if args and isinstance(args[0], Iterable):
        kwargs = _args2kwargs(args, columns)
    elif args:
        raise TypeError(f"Invalid args: {args}")

    if kwargs:
        return _parse_kwargs(kwargs, columns)
    return {}, 0


def _check_missing_columns(array_columns: tuple[str, ...], defaults: dict[str, np.generic], provided_columns: set[str]):
    required_columns = set(array_columns) - set(defaults.keys())
    if missing_columns := required_columns - provided_columns:
        raise ValueError(f"Missing required columns: {missing_columns}")


def _fill_defaults(array: np.ndarray, defaults: dict[str, np.generic]):
    """Fills the defaults into the array."""
    for column, default in defaults.items():
        if default is empty:
            column_type: type = array.dtype[column]
            array[column] = empty(column_type)  # type: ignore[call-overload]
        else:
            array[column] = default  # type: ignore[call-overload]


def _fill_with_kwargs(array: np.ndarray, kwargs: dict[str, np.ndarray]):
    """Fills the kwargs into the array."""
    for column, values in kwargs.items():
        array[column] = values  # type: ignore[call-overload]


def _parse_structured_array(from_array: np.ndarray, to_array: np.ndarray) -> np.ndarray:
    shared_columns, ignored_columns = _determine_column_overlap(from_array, to_array)
    if ignored_columns:
        logging.debug(
            "Ignored provided columns %s during build of array with columns %s", ignored_columns, to_array.dtype.names
        )
    to_array[shared_columns] = from_array[shared_columns]  # type: ignore[index]
    return to_array


def _determine_column_overlap(from_array: np.ndarray, to_array: np.ndarray) -> tuple[list[str], list[str]]:
    """Returns two lists: columns present in both arrays and the columns that are only present in from_array"""
    from_columns = set(from_array.dtype.names or ())
    to_columns = set(to_array.dtype.names or ())

    return list(from_columns & to_columns), list(from_columns - to_columns)


def _parse_array(array: np.ndarray, dtype: np.dtype):
    if len(array.shape) == 1 and array.dtype == dtype:
        return array
    if len(array.shape) == 1:
        return np.array(array, dtype=dtype)
    if len(array.shape) == 2:
        return _parse_2d_array(array, dtype)
    raise NotImplementedError(f"Unsupported array shape {array.shape}")


def _parse_2d_array(array: np.ndarray, dtype: np.dtype):
    """Parses the 2d array to a 1d array."""
    columns: list[str] = list(dtype.names) if dtype.names else []
    if len(columns) not in array.shape:
        raise ValueError(f"Cannot convert array of shape {array.shape} into {len(columns)} columns.")
    column_dim = 0 if len(columns) == array.shape[0] else 1
    size_dim = 1 if column_dim == 0 else 0
    new_array = np.ones(array.shape[size_dim], dtype=dtype)
    for index, column in enumerate(columns):
        if column_dim == 0:
            new_array[column] = array[index, :]  # type: ignore[call-overload]
        else:
            new_array[column] = array[:, index]  # type: ignore[call-overload]
    return new_array


def _parse_kwargs(kwargs: dict[str, list | np.ndarray], columns: list[str]) -> tuple[dict[str, np.ndarray], int]:
    """Parses the kwargs to a dict of np.ndarrays."""
    parsed_kwargs = {}

    size = 0
    for column, values in kwargs.items():
        parsed_kwargs[column] = np.array(values).flatten()

        value_size = _get_size(values)
        if size == 0:
            size = value_size
        elif size != len(values):
            raise ValueError(f"Size of column '{column}' does not match other columns.")

    if invalid_columns := set(parsed_kwargs.keys()) - set(columns):
        raise ValueError(f"Invalid columns: {invalid_columns}")
    return parsed_kwargs, size


def _get_size(values: Sized):
    """Returns the size of the values."""
    if is_sequence(values):
        return len(values)
    return 1


def _args2kwargs(args: tuple[Any, ...], columns: list[str]) -> dict[str, list]:
    """Parses the args to kwargs."""
    kwargs = {}
    if len(args) == 1:
        args = args[0]

    args_as_array = np.array(args)
    if len(args_as_array.shape) != 2:
        raise ValueError(
            "Cannot parse args: input is not 2D, probably due to an inconsistent number of values per row."
        )

    _, args_n_columns = args_as_array.shape
    if args_n_columns != len(columns):
        raise ValueError(f"Cannot parse args: requires {len(columns)} columns per row, got {args_n_columns}.")

    for index, column in enumerate(columns):
        kwargs[column] = [row[index] for row in args]
    return kwargs



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0
from typing import Any

from dash import Input, Output, callback

from power_grid_model_ds._core.visualizer.layout.colors import CYTO_COLORS
from power_grid_model_ds._core.visualizer.layout.cytoscape_styling import DEFAULT_STYLESHEET


@callback(
    Output("cytoscape-graph", "stylesheet"),
    Input("search-form-group-input", "value"),
    Input("search-form-column-input", "value"),
    Input("search-form-operator-input", "value"),
    Input("search-form-value-input", "value"),
)
def search_element(group: str, column: str, operator: str, value: str) -> list[dict[str, Any]]:
    """Color the specified element red based on the input values."""
    if not group or not column or not value:
        return DEFAULT_STYLESHEET

    # Determine if we're working with a node or an edge type
    if group == "node":
        style = {
            "background-color": CYTO_COLORS["highlighted"],
            "text-background-color": CYTO_COLORS["highlighted"],
        }
    else:
        style = {"line-color": CYTO_COLORS["highlighted"], "target-arrow-color": CYTO_COLORS["highlighted"]}

    if column == "id":
        selector = f'[{column} {operator} "{value}"]'
    else:
        selector = f"[{column} {operator} {value}]"

    new_style = {
        "selector": selector,
        "style": style,
    }
    return DEFAULT_STYLESHEET + [new_style]


@callback(
    Output("search-form-column-input", "options"),
    Output("search-form-column-input", "value"),
    Input("search-form-group-input", "value"),
    Input("columns-store", "data"),
)
def update_column_options(selected_group, store_data):
    """Update the column dropdown options based on the selected group."""
    if not selected_group or not store_data:
        return [], None

    # Get columns for the selected group (node, line, link, or transformer)
    columns = store_data.get(selected_group, [])
    default_value = columns[0] if columns else "id"

    return columns, default_value



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Helper functions for (un)zipping files"""

import gzip
import logging
import os
import shutil
from pathlib import Path
from typing import List
from zipfile import ZipFile

_logger = logging.getLogger(__name__)


def zip_folder(folder_to_zip: Path) -> Path:
    """Zip folder contents"""
    zip_file_path = folder_to_zip.with_suffix(".zip").resolve()
    folder_to_zip = folder_to_zip.resolve()

    # move into folder to zip to avoid nested folders in .zip file
    current_working_directory = os.getcwd()
    os.chdir(folder_to_zip)
    files_to_zip = [path.relative_to(folder_to_zip) for path in folder_to_zip.rglob("*")]
    zip_files(zip_file_path, files_to_zip)

    # move out of folder again
    os.chdir(current_working_directory)

    return zip_file_path


def zip_files(zip_file_path: Path, files_to_zip: List[Path]) -> Path:
    """Zip files"""
    with ZipFile(str(zip_file_path.with_suffix(".zip")), "w") as zip_object:
        for file in files_to_zip:
            zip_object.write(str(file))
        zip_object.close()
    return zip_file_path.resolve()


def unzip_files(zip_file_path: Path) -> Path:
    """Unzip .zip file to folder"""
    extraction_dir = zip_file_path.with_suffix("")
    extraction_dir.mkdir(parents=True, exist_ok=True)
    with ZipFile(str(zip_file_path.resolve()), "r") as zip_ref:
        zip_ref.extractall(extraction_dir)
    return extraction_dir


def gzip2file(gzip_path: Path) -> Path:
    """unzip a gzip (.gz) file"""
    _logger.info(f"Unzipping {gzip_path.name}")

    file_path = gzip_path.with_suffix("")
    with gzip.open(gzip_path, "rb") as f_in:
        with open(file_path, "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)
    return file_path


def file2gzip(file_path: Path) -> Path:
    """zip a gzip (.gz) file"""
    _logger.info(f"Zipping {file_path.name}")

    gzip_path = file_path.with_suffix(f"{file_path.suffix}.gz")
    with open(file_path, "rb") as f_in:
        with gzip.open(gzip_path, "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)
    return gzip_path



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Base generator"""

import numpy as np

from power_grid_model_ds._core.model.grids.base import Grid


class BaseGenerator:
    """Base class to build a generator for grid elements"""

    def __init__(self, grid: Grid, seed: int) -> None:
        """Initializes generator with grid and amount"""
        self.grid = grid

        self.starting_seed = seed
        self.rng = np.random.default_rng(seed)

    def reset_rng(self, seed: int):
        """Sets the rng for a generator"""
        rng = np.random.default_rng(seed)
        self.rng = rng



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from abc import ABC
from collections import namedtuple
from copy import copy
from functools import lru_cache
from typing import Any, Iterable, Literal, Type, TypeVar

import numpy as np
from numpy.typing import ArrayLike, NDArray

from power_grid_model_ds._core.model.arrays.base._build import build_array
from power_grid_model_ds._core.model.arrays.base._filters import apply_exclude, apply_filter, apply_get, get_filter_mask
from power_grid_model_ds._core.model.arrays.base._modify import check_ids, re_order, update_by_id
from power_grid_model_ds._core.model.arrays.base._optional import pandas
from power_grid_model_ds._core.model.arrays.base._string import convert_array_to_string
from power_grid_model_ds._core.model.arrays.base.errors import ArrayDefinitionError
from power_grid_model_ds._core.model.constants import EMPTY_ID, empty
from power_grid_model_ds._core.utils.misc import get_inherited_attrs

# pylint: disable=missing-function-docstring, too-many-public-methods

_RESERVED_COLUMN_NAMES: set = set(dir(np.array([]))).union({"data"})
_DEFAULT_STR_LENGTH: int = 50

Column = NDArray

Self = TypeVar("Self", bound="FancyArray")


class FancyArray(ABC):
    """Base class for all arrays.

    You can create your own array by subclassing FancyArray.
    Array-columns can be defined by adding class attributes with the column name and the numpy dtype.

    Example:
        >>> class MyArray(FancyArray):
        >>>     id: NDArray[np.int64]
        >>>     name: NDArray[np.str_]
        >>>     value: NDArray[np.float64]

    Note on string-columns:
        The default length for string columns is stored in _DEFAULT_STR_LENGTH.
        To change this, you can set the _str_lengths class attribute.

    Example:
        >>> class MyArray(FancyArray):
        >>>     name: NDArray[np.str_]
        >>>     _str_lengths = {"name": 100}

    Extra note on string-columns:
        Where possible, it is recommended use IntEnum's instead of string-columns to reduce memory usage.
    """

    _data: NDArray = np.ndarray([])
    _defaults: dict[str, Any] = {}
    _str_lengths: dict[str, int] = {}

    def __init__(self: Self, *args, data: NDArray | None = None, **kwargs):
        if data is None:
            self._data = build_array(*args, dtype=self.get_dtype(), defaults=self.get_defaults(), **kwargs)
        else:
            self._data = data

    @property
    def data(self: Self):
        return self._data

    @classmethod
    @lru_cache
    def get_defaults(cls) -> dict[str, Any]:
        return get_inherited_attrs(cls, "_defaults")["_defaults"]

    @classmethod
    @lru_cache
    def get_dtype(cls):
        annotations = get_inherited_attrs(cls, "_str_lengths")
        str_lengths = annotations.pop("_str_lengths")
        dtypes = {}
        for name, dtype in annotations.items():
            if len(dtype.__args__) > 1:
                # regular numpy dtype (i.e. without shape)
                dtypes[name] = dtype.__args__[1].__args__[0]
            elif hasattr(dtype, "__metadata__"):
                # metadata annotation contains shape
                # define dtype using a (type, shape) tuple
                # see: #1 in https://numpy.org/doc/stable/user/basics.rec.html#structured-datatype-creation
                dtype_type = dtype.__args__[0].__args__[1].__args__[0]
                dtype_shape = dtype.__metadata__[0].__args__
                dtypes[name] = (dtype_type, dtype_shape)
            else:
                raise ValueError(f"dtype {dtype} not understood or supported")

        if not dtypes:
            raise ArrayDefinitionError("Array has no defined Columns")
        if reserved := set(dtypes.keys()) & _RESERVED_COLUMN_NAMES:
            raise ArrayDefinitionError(f"Columns cannot be reserved names: {reserved}")

        dtype_list = []
        for name, dtype in dtypes.items():
            if dtype is np.str_:
                string_length = str_lengths.get(name, _DEFAULT_STR_LENGTH)
                dtype_list.append((name, np.dtype(f"U{string_length}")))
            elif dtype is tuple:
                dtype_list.append((name, *dtype))
            else:
                dtype_list.append((name, dtype))
        return np.dtype(dtype_list)

    def __repr__(self: Self) -> str:
        try:
            data = getattr(self, "data")
            if data.size > 3:
                return f"{self.__class__.__name__}([{data[:3]}]... + {data.size - 3} more rows)"
            return f"{self.__class__.__name__}([{data}])"
        except AttributeError:
            return self.__class__.__name__ + "()"

    def __str__(self) -> str:
        return self.as_table()

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self: Self):
        for record in self._data:
            yield self.__class__(data=np.array([record]))

    def __getattr__(self: Self, attr):
        if attr == "__array_interface__":
            # prevent unintended usage of numpy functions. np.unique/np.sort give wrong results.
            raise TypeError(
                "Cannot use numpy functions directly on FancyArray. "
                "Instead, you can use (fancypy) fp.unique/fp.sort or np.unique/np.sort on array.data"
            )
        if attr.startswith("__"):  # prevent unintended usage of numpy magic methods
            raise AttributeError(f"Cannot get attribute {attr} on {self.__class__.__name__}")

        if attr in self.get_dtype().names:
            return self._data[attr]
        return getattr(self._data, attr)

    def __setattr__(self: Self, attr: str, value: object) -> None:
        if attr in ["_data", "_defaults"]:
            super().__setattr__(attr, value)
            return
        try:
            self._data[attr] = value  # type: ignore[call-overload]
        except (AttributeError, ValueError) as error:
            raise AttributeError(f"Cannot set attribute {attr} on {self.__class__.__name__}") from error

    def __getitem__(self: Self, item):
        """Used by for-loops, slicing [0:3], column-access ['id'], row-access [0], multi-column access.
        Note: If a single item is requested, return a named tuple instead of a np.void object.
        """

        result = self._data.__getitem__(item)

        if isinstance(item, (list, tuple)) and (len(item) == 0 or np.array(item).dtype.type is np.bool_):
            return self.__class__(data=result)
        if isinstance(item, (str, list, tuple)):
            return result
        if isinstance(result, np.void):
            return self.__class__(data=np.array([result]))
        return self.__class__(data=result)

    def __setitem__(self: Self, key, value):
        if isinstance(value, FancyArray):
            value = value.data
        return self._data.__setitem__(key, value)

    def __contains__(self: Self, item: Self) -> bool:
        if isinstance(item, FancyArray):
            return item.data in self._data
        return False

    def __hash__(self: Self):
        return hash(f"{self.__class__} {self}")

    def __eq__(self: Self, other):
        return self._data.__eq__(other.data)

    def __copy__(self: Self):
        return self.__class__(data=copy(self._data))

    def copy(self: Self):
        """Return a copy of this array including its data"""
        return copy(self)

    @classmethod
    def zeros(cls: Type[Self], num: int, empty_id: bool = True) -> Self:
        """Construct an array filled with zeros.

        If 'empty_id' is True, the 'id' column will be filled with EMPTY_ID values.
        """
        dtype = cls.get_dtype()
        zeros_array = np.zeros(num, dtype=dtype)
        if empty_id and "id" in dtype.names:
            zeros_array["id"] = EMPTY_ID  # type: ignore[call-overload]
        return cls(data=zeros_array)

    @classmethod
    def empty(cls: Type[Self], num: int, use_defaults: bool = True) -> Self:
        """Construct an array filled with 'empty' values.

        'empty' values differs per dtype:
            string: "" (empty string)
            float: np.nan
            integer: minimum value

        if 'use_defaults' is True, the default values will be used instead where possible.
        """
        array_dtype = cls.get_dtype()
        array = np.zeros(num, dtype=array_dtype)

        defaults = cls.get_defaults()
        for column in array_dtype.names:
            if use_defaults and column in defaults and defaults[column] is not empty:
                array[column] = defaults[column]
            else:
                array[column] = empty(array_dtype[column])
        return cls(data=array)

    def is_empty(self, column: str) -> NDArray[np.bool_]:
        """Check if a column is filled with 'empty' values."""
        empty_value = self.get_empty_value(column)
        if empty_value is np.nan:
            return np.isnan(self._data[column])
        return np.isin(self._data[column], empty_value)

    def get_empty_value(self, column: str) -> float | int | str | bool:
        array_dtype = self.get_dtype()
        return empty(array_dtype[column])

    def set_empty(self, column: str):
        """Set a column to its 'empty' value."""
        array_dtype = self.get_dtype()
        self._data[column] = empty(array_dtype[column])  # type: ignore[call-overload]

    @property
    def columns(self) -> list[str]:
        return list(self.get_dtype().names)

    @property
    def record(self):
        """Return a named tuple of the first record in the array."""
        if self.size != 1:
            raise ValueError(f"Cannot return record of array with size {self.size}")

        class_name = self.__class__.__name__
        tpl_cls = namedtuple(f"{class_name}Record", self.dtype.names)
        if isinstance(self._data, np.void):
            return tpl_cls(*self._data)
        return tpl_cls(*self._data[0])

    def filter(
        self: Self,
        *args: int | Iterable[int] | np.ndarray,
        mode_: Literal["AND", "OR"] = "AND",
        **kwargs: Any | list[Any] | np.ndarray,
    ) -> Self:
        return self.__class__(data=apply_filter(*args, array=self._data, mode_=mode_, **kwargs))

    def exclude(
        self: Self,
        *args: int | Iterable[int] | np.ndarray,
        mode_: Literal["AND", "OR"] = "AND",
        **kwargs: Any | list[Any] | np.ndarray,
    ) -> Self:
        return self.__class__(data=apply_exclude(*args, array=self._data, mode_=mode_, **kwargs))

    def get(
        self: Self,
        *args: int | Iterable[int] | np.ndarray,
        mode_: Literal["AND", "OR"] = "AND",
        **kwargs: Any | list[Any] | np.ndarray,
    ) -> Self:
        return self.__class__(data=apply_get(*args, array=self._data, mode_=mode_, **kwargs))

    def filter_mask(
        self: Self,
        *args: int | Iterable[int] | np.ndarray,
        mode_: Literal["AND", "OR"] = "AND",
        **kwargs: Any | list[Any] | np.ndarray,
    ) -> np.ndarray:
        return get_filter_mask(*args, array=self._data, mode_=mode_, **kwargs)

    def exclude_mask(
        self: Self,
        *args: int | Iterable[int] | np.ndarray,
        mode_: Literal["AND", "OR"] = "AND",
        **kwargs: Any | list[Any] | np.ndarray,
    ) -> np.ndarray:
        return ~get_filter_mask(*args, array=self._data, mode_=mode_, **kwargs)

    def re_order(self: Self, new_order: ArrayLike, column: str = "id") -> Self:
        return self.__class__(data=re_order(self._data, new_order, column=column))

    def update_by_id(self: Self, ids: ArrayLike, allow_missing: bool = False, **kwargs) -> None:
        try:
            _ = update_by_id(self._data, ids, allow_missing, **kwargs)
        except ValueError as error:
            raise ValueError(f"Cannot update {self.__class__.__name__}. {error}") from error

    def get_updated_by_id(self: Self, ids: ArrayLike, allow_missing: bool = False, **kwargs) -> Self:
        try:
            mask = update_by_id(self._data, ids, allow_missing, **kwargs)
            return self.__class__(data=self._data[mask])
        except ValueError as error:
            raise ValueError(f"Cannot update {self.__class__.__name__}. {error}") from error

    def check_ids(self: Self, return_duplicates: bool = False) -> NDArray | None:
        return check_ids(self._data, return_duplicates=return_duplicates)

    def as_table(self: Self, column_width: int | str = "auto", rows: int = 10) -> str:
        return convert_array_to_string(self, column_width=column_width, rows=rows)

    def as_df(self: Self):
        """Convert to pandas DataFrame"""
        if pandas is None:
            raise ImportError("pandas is not installed")
        return pandas.DataFrame(self._data)

    @classmethod
    def from_extended(cls: Type[Self], extended: Self) -> Self:
        """Create an instance from an extended array."""
        if not isinstance(extended, cls):
            raise TypeError(f"Extended array must be of type {cls.__name__}, got {type(extended).__name__}")
        dtype = cls.get_dtype()
        return cls(data=np.array(extended[list(dtype.names)], dtype=dtype))



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""Generator for SourceArray"""

import numpy as np

from power_grid_model_ds._core.data_source.generator.arrays.base import BaseGenerator
from power_grid_model_ds._core.model.arrays import NodeArray, SourceArray
from power_grid_model_ds._core.model.enums.nodes import NodeType


class SourceGenerator(BaseGenerator):
    """Generator for source elements in the grid (substations)"""

    def run(self, amount: int) -> tuple[NodeArray, SourceArray]:
        """Generate nodes in a grid which are sources (substations)"""
        substation_node_array = self.grid.node.__class__.empty(amount)
        substation_node_array.id = 1 + self.grid.max_id + np.arange(amount)
        substation_node_array.u_rated = 10_500
        substation_node_array.node_type = NodeType.SUBSTATION_NODE.value

        source_array = self.grid.source.__class__.empty(amount)
        source_array.id = 1 + substation_node_array.id.max() + np.arange(amount)
        source_array.node = substation_node_array.id
        source_array.status = 1
        source_array.u_ref = 1

        return substation_node_array, source_array
